%Author: Shebuti Rayana

%input:
% X - N X d data matrix, N = number of data points, d = number of dimensions
% y - N X 1 label vector, 1 = outliers, 0 = inliers
% k - value of k for k-nearest neighbor distance e.g., k = 5
% base_algo - name of the base algorithm, supports 'AvgKNN' (average
% k-nearesr neighbor) and 'LOF' (local outlier factor)
% iter - maximum number of iterations if stopping condition does not meet 

%output:
% fs - score vector for all data points
% r - rank list for all data points (most to least outlierness score)
% ret_iter - iteration at which the final result is settled
% ap - average precision of the final result

%Required s/w to run this code: CVX (Matlab Software for Disciplined Convex Programming)
% can be downloaded from http://cvxr.com/cvx/

%For the AR optimization problem read following paper:
%A. Platanios, A. Blum, and T. M. Mitchell, ?Estimating accuracy from
%unlabeled data,? in In Proceedings of UAI, 2014.

% Copyright: Shebuti Rayana
% PhD student, Computer Science Department, Stony Brook University
% Stony Brook, NY - 11794

function [fs,r,ret_iter,ap] = CARE(X,y,k,base_algo,iter)
    [N,d] = size(X);
    pbc = 100; %number of parallel base components using feature bagging
    
    n = nchoosek(pbc,2) + pbc; % number of singles and pairs of base detectors
    m = nchoosek(pbc,2); % number of pairs of base detectors 
    W = Constraints100(n,m); % coefficients for AR optimization
    
    points = [1:N]'; % initially run the base detectors for all data points
    t = 0;
    save_score = [];
    auc = [];
    sd = [];
    Ave_AUC = [];
    ex_vector = [];
    
    while t <= iter
        scores = [];
        for i = 1:pbc
            sample = randi([floor(d/2) d-1],1,1);
            ind = sort(randsample(1:d,sample)); % random feature selection
            if strcmp(base_algo,'LOF')
                sc = LOF(X(:,ind), points, k);
            end
            if strcmp(base_algo,'AvgKNN')
                sc = AvgKNN(X(:,ind), points, k);
            end
            scores = [scores,sc];
        end
        
        per = 0.2; % find agreement for top 20% points using Cantelli's inequality
        kk1 = sqrt((1-per)/per);
        th = mean(scores,1)+kk1*std(scores); % threshold
        
        Rank = [];
        for i = 1:pbc
            Rank = union(Rank,find(scores(:,i) >= th(i)));
        end
        
        R = length(Rank);
        
        C = [];
        class = zeros(N,1);
        for i = 1:pbc
            C0 = class;
            C0(scores(:,i) >= th(i)) = 1;
            C = [C, C0];
        end

        A = ConstraintsAgreement100(C, Rank, m); % calculating agreement for AR optimization
        
        % stopping criteria based on A (pairwise agreement rates)
        [f,x] = hist(A,[0:0.01:1]);
        temp = cumsum(f./sum(f)); 
        auc = [auc;trapz(x,1-temp')]; % AUC of ccdf of A
        ex_vector = [ex_vector;t+1];
        Ave_AUC = cumsum(auc)./ex_vector;
        
        if t > 0
            sumX = sum(auc,1);
            sumX2 = sum(auc.^2,1);
            sd = [sd;sqrt((sumX2-(sumX.^2)./(t+1))./(t))];
        else
            sd = [sd;auc(t+1)];
        end
        
        if auc(t+1) - Ave_AUC(t+1) >= 0 || abs(auc(t+1) - Ave_AUC(t+1)) <= sd(t+1)
            % AR optimization to calculate single and pairwise errors
            cvx_begin
                variables e(n,1)
                variable epsilon(n,1)
                minimize(sum(square(e))+sum(epsilon))
                subject to
                   A == W*e + 1;
                   0 <= e <= 0.5 + epsilon
                   0 <= epsilon 
            cvx_end

            e(e == 0) = 10^-6; % to avoid divide by zero error in calculating weight
            weight = log((2./e(1:pbc) - 1))./2; % calculate weight for each base detector

            % unify the outlierness scores in [0,1]
            Pr = [];
            for i = 1:pbc
                 Pr0 = Unification(scores(:,i), 'Gaussian');
                 Pr = [Pr, Pr0];
            end

            ind = find(e(1:pbc) < 0.5); % remove the detectors having error >= 0.5

            % if all detectors have error >= 0.5 find a cluster of detectors
            % having comparatively lower error (this situation does not occur frequently)

            if isempty(ind)
                %clustering approach
                centroid = [min(e(1:pbc));max(e(1:pbc))]; %initialize centroid
                [IDX,C] = kmeans(e(1:pbc),2,'start',centroid);

                if C(2) > C(1)
                    ind = find(IDX == 1);
                else
                    if C(1) > C(2)
                        ind = find(IDX == 2);
                    end
                end
            end

            score = (Pr(:,ind)*weight(ind))./sum(weight(ind)); % weighted summation
            score = (score - min(score))./(max(score) - min(score)); % normalizing final score
            save_score = [save_score,score];
            mean_score = mean(save_score, 2);

            points = FVPS(mean_score, N); % filtered variable probability sampling

            t = t + 1;
            ret_iter = t;
        else
            if abs(auc(t+1) - Ave_AUC(t+1)) > sd(t+1)
                %stop_iter = t+1;
                ret_iter = t;
                break;
            end
        end
    end
    
    % final result
    if ret_iter == 0
        fs = save_score(:,1);
    else
        fs = mean(save_score(:,1:ret_iter),2);
    end
    [~,r] = sort(fs,'descend');
    ap = AveragePrecision(r,find(y == 1),0);
end

%% Filteres variable probability sampling
function [ index ] = FVPS(score, N)
    per = 0.2; % find agreement for top 20% points using Cantelli's inequality
    kk1 = sqrt((1-per)/per);
    th = mean(score,1)+kk1*std(score); % threshold for filtering top 20%
    
    K = sum(score >= th);
    st = min(1,50/N);
    en = min(1,1000/N);
    f = unifrnd(st,en);
    Sample_size = floor((N - K) * f); 
        
    [~,R] = sort(score,'descend');
    topR = R(1:K);
    score = 1 - score; % normality score
    score = score./sum(score);
    n = length(score);
    index = zeros(Sample_size, 1);
    [Ascore,rank] = sort(score, 'ascend');
    
    for i = 2:n
        Ascore(i) = Ascore(i) + Ascore(i-1);
    end
    
    randNum = rand(1,Sample_size) ;
    for i = 1:Sample_size
        j = 1;
        while j <= n
            if randNum(i) <= Ascore(j)
                if ~ismember(rank(j),topR)
                    index(i,1) = rank(j);
                    topR = [topR;rank(j)];
                    break;
                else
                    randNum(i) = rand;
                    j = 0;
                end
            end
            j = j+1;
        end
        if index(i,1) == 0
            index(i, 1) = rank(n);
        end
    end
end
